@@ -22,10 +22,11 @@ import (
 	apierrors "k8s.io/apimachinery/pkg/api/errors"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
-	"k8s.io/apimachinery/pkg/labels"
+	"k8s.io/apimachinery/pkg/fields"
 	"k8s.io/apimachinery/pkg/runtime"
 	"k8s.io/apimachinery/pkg/runtime/schema"
 	k8sjson "k8s.io/apimachinery/pkg/runtime/serializer/json"
+	"k8s.io/apimachinery/pkg/types"
 	"k8s.io/apimachinery/pkg/util/diff"
 	"k8s.io/apimachinery/pkg/util/wait"
 	"k8s.io/apimachinery/pkg/watch"
@@ -33,12 +34,11 @@ import (
 	"k8s.io/client-go/dynamic"
 	k8sscheme "k8s.io/client-go/kubernetes/scheme"
 	"k8s.io/client-go/rest"
-	"k8s.io/component-base/featuregate"
+	k8scontrollerclient "sigs.k8s.io/controller-runtime/pkg/client"
 
 	"github.com/operator-framework/api/pkg/operators/v1alpha1"
 	"github.com/operator-framework/operator-lifecycle-manager/pkg/api/client/clientset/versioned"
 	"github.com/operator-framework/operator-lifecycle-manager/pkg/controller/registry"
-	"github.com/operator-framework/operator-lifecycle-manager/pkg/feature"
 	controllerclient "github.com/operator-framework/operator-lifecycle-manager/pkg/lib/controller-runtime/client"
 	"github.com/operator-framework/operator-lifecycle-manager/pkg/lib/operatorclient"
 	pmversioned "github.com/operator-framework/operator-lifecycle-manager/pkg/package-server/client/clientset/versioned"
@@ -50,49 +50,16 @@ const (
 	pollInterval = 1 * time.Second
 	pollDuration = 5 * time.Minute
 
-	olmConfigMap = "olm-operators"
+	olmConfigMap = "olm-operators" // No-longer used, how long do we keep this around?
+
 	// sync name with scripts/install_local.sh
 	packageServerCSV = "packageserver.v1.0.0"
 )
 
 var (
-	cleaner *namespaceCleaner
 	genName = names.SimpleNameGenerator.GenerateName
-
-	persistentCatalogNames               = []string{olmConfigMap}
-	nonPersistentCatalogsFieldSelector   = createFieldNotEqualSelector("metadata.name", persistentCatalogNames...)
-	persistentConfigMapNames             = []string{olmConfigMap}
-	nonPersistentConfigMapsFieldSelector = createFieldNotEqualSelector("metadata.name", persistentConfigMapNames...)
-	persistentCSVNames                   = []string{packageServerCSV}
-	nonPersistentCSVFieldSelector        = createFieldNotEqualSelector("metadata.name", persistentCSVNames...)
 )
 
-type namespaceCleaner struct {
-	namespace      string
-	skipCleanupOLM bool
-}
-
-func newNamespaceCleaner(namespace string) *namespaceCleaner {
-	return &namespaceCleaner{
-		namespace:      namespace,
-		skipCleanupOLM: false,
-	}
-}
-
-// notifyOnFailure checks if a test has failed or cleanup is true before cleaning a namespace
-func (c *namespaceCleaner) NotifyTestComplete(cleanup bool) {
-	if CurrentGinkgoTestDescription().Failed {
-		c.skipCleanupOLM = true
-	}
-
-	if c.skipCleanupOLM || !cleanup {
-		ctx.Ctx().Logf("skipping cleanup")
-		return
-	}
-
-	cleanupOLM(c.namespace)
-}
-
 // newKubeClient configures a client to talk to the cluster defined by KUBECONFIG
 func newKubeClient() operatorclient.ClientInterface {
 	return ctx.Ctx().KubeClient()
@@ -344,11 +311,15 @@ func registryPodHealthy(address string) bool {
 func catalogSourceRegistryPodSynced(catalog *v1alpha1.CatalogSource) bool {
 	registry := catalog.Status.RegistryServiceStatus
 	connState := catalog.Status.GRPCConnectionState
-	if registry != nil && connState != nil && !connState.LastConnectTime.IsZero() {
+	if registry != nil && connState != nil && !connState.LastConnectTime.IsZero() && connState.LastObservedState == "READY" {
 		fmt.Printf("catalog %s pod with address %s\n", catalog.GetName(), registry.Address())
 		return registryPodHealthy(registry.Address())
 	}
-	fmt.Printf("waiting for catalog pod %v to be available (for sync)\n", catalog.GetName())
+	state := "NO_CONNECTION"
+	if connState != nil {
+		state = connState.LastObservedState
+	}
+	fmt.Printf("waiting for catalog pod %v to be available (for sync) - %s\n", catalog.GetName(), state)
 	return false
 }
 
@@ -368,74 +339,163 @@ func fetchCatalogSourceOnStatus(crc versioned.Interface, name, namespace string,
 	return fetched, err
 }
 
-func createFieldNotEqualSelector(field string, names ...string) string {
+// createFieldNotEqualSelector generates a field selector that matches resources that have a field value that DOES NOT match any of a set of values.
+// This function panics if the generated selector cannot be parsed.
+func createFieldNotEqualSelector(field string, values ...string) fields.Selector {
 	var builder strings.Builder
-	for i, name := range names {
+	for i, value := range values {
 		builder.WriteString(field)
 		builder.WriteString("!=")
-		builder.WriteString(name)
-		if i < len(names)-1 {
+		builder.WriteString(value)
+		if i < len(values)-1 {
 			builder.WriteString(",")
 		}
 	}
 
-	return builder.String()
+	selector, err := fields.ParseSelector(builder.String())
+	if err != nil {
+		panic(fmt.Errorf("failed to build fields-not-equal selector: %s", err))
+	}
+
+	return selector
 }
 
-func cleanupOLM(namespace string) {
-	var immediate int64 = 0
-	crc := newCRClient()
-	c := newKubeClient()
+// MaskNotFound "masks" an given error by returning nil when it refers to a "NotFound" API status response, otherwise returns the error unaltered.
+func MaskNotFound(err error) error {
+	if apierrors.IsNotFound(err) {
+		return nil
+	}
+
+	return err
+}
+
+var (
+	persistentCatalogNames        = []string{olmConfigMap}
+	ephemeralCatalogFieldSelector = k8scontrollerclient.MatchingFieldsSelector{
+		Selector: createFieldNotEqualSelector("metadata.name", persistentCatalogNames...),
+	}
+	persistentConfigMapNames         = []string{olmConfigMap}
+	ephemeralConfigMapsFieldSelector = k8scontrollerclient.MatchingFieldsSelector{
+		Selector: createFieldNotEqualSelector("metadata.name", persistentConfigMapNames...),
+	}
+	persistentCSVNames        = []string{packageServerCSV}
+	ephemeralCSVFieldSelector = k8scontrollerclient.MatchingFieldsSelector{
+		Selector: createFieldNotEqualSelector("metadata.name", persistentCSVNames...),
+	}
+)
+
+// TearDown deletes all OLM resources in the corresponding namespace and at the cluster scope.
+func TearDown(namespace string) {
+	var (
+		clientCtx   = context.Background()
+		client      = ctx.Ctx().Client()
+		inNamespace = k8scontrollerclient.InNamespace(namespace)
+		logf        = ctx.Ctx().Logf
+	)
 
 	// Cleanup non persistent OLM CRs
-	ctx.Ctx().Logf("cleaning up any remaining non persistent resources...")
-	deleteOptions := metav1.DeleteOptions{GracePeriodSeconds: &immediate}
-	listOptions := metav1.ListOptions{}
+	logf("cleaning up ephemeral test resources...")
 
-	err := crc.OperatorsV1alpha1().ClusterServiceVersions(namespace).DeleteCollection(context.TODO(), deleteOptions, metav1.ListOptions{FieldSelector: nonPersistentCSVFieldSelector})
-	Expect(err).NotTo(HaveOccurred())
+	logf("deleting test subscriptions...")
+	Eventually(func() error {
+		return client.DeleteAllOf(clientCtx, &v1alpha1.Subscription{}, inNamespace)
+	}).Should(Succeed(), "failed to delete test subscriptions")
 
-	err = crc.OperatorsV1alpha1().InstallPlans(namespace).DeleteCollection(context.TODO(), deleteOptions, listOptions)
-	Expect(err).NotTo(HaveOccurred())
+	Eventually(func() (remaining []v1alpha1.Subscription, err error) {
+		list := &v1alpha1.SubscriptionList{}
+		err = client.List(clientCtx, list, inNamespace)
+		if list != nil {
+			remaining = list.Items
+		}
 
-	err = crc.OperatorsV1alpha1().Subscriptions(namespace).DeleteCollection(context.TODO(), deleteOptions, listOptions)
-	Expect(err).NotTo(HaveOccurred())
+		return
+	}).Should(BeEmpty(), "failed to await deletion of test subscriptions")
 
-	err = crc.OperatorsV1alpha1().CatalogSources(namespace).DeleteCollection(context.TODO(), deleteOptions, metav1.ListOptions{FieldSelector: nonPersistentCatalogsFieldSelector})
-	Expect(err).NotTo(HaveOccurred())
+	logf("deleting test installplans...")
+	Eventually(func() error {
+		return client.DeleteAllOf(clientCtx, &v1alpha1.InstallPlan{}, inNamespace)
+	}).Should(Succeed(), "failed to delete test installplans")
 
-	// error: the server does not allow this method on the requested resource
-	// Cleanup non persistent configmaps
-	err = c.KubernetesInterface().CoreV1().Pods(namespace).DeleteCollection(context.TODO(), deleteOptions, metav1.ListOptions{})
-	Expect(err).NotTo(HaveOccurred())
+	Eventually(func() (remaining []v1alpha1.InstallPlan, err error) {
+		list := &v1alpha1.InstallPlanList{}
+		err = client.List(clientCtx, list, inNamespace)
+		if list != nil {
+			remaining = list.Items
+		}
 
-	err = waitForEmptyList(func() (int, error) {
-		res, err := crc.OperatorsV1alpha1().ClusterServiceVersions(namespace).List(context.TODO(), metav1.ListOptions{FieldSelector: nonPersistentCSVFieldSelector})
-		ctx.Ctx().Logf("%d %s remaining", len(res.Items), "csvs")
-		return len(res.Items), err
-	})
-	Expect(err).NotTo(HaveOccurred())
+		return
+	}).Should(BeEmpty(), "failed to await deletion of test installplans")
 
-	err = waitForEmptyList(func() (int, error) {
-		res, err := crc.OperatorsV1alpha1().InstallPlans(namespace).List(context.TODO(), metav1.ListOptions{})
-		ctx.Ctx().Logf("%d %s remaining", len(res.Items), "installplans")
-		return len(res.Items), err
-	})
-	Expect(err).NotTo(HaveOccurred())
+	logf("deleting test catalogsources...")
+	Eventually(func() error {
+		return client.DeleteAllOf(clientCtx, &v1alpha1.CatalogSource{}, inNamespace, ephemeralCatalogFieldSelector)
+	}).Should(Succeed(), "failed to delete test catalogsources")
 
-	err = waitForEmptyList(func() (int, error) {
-		res, err := crc.OperatorsV1alpha1().Subscriptions(namespace).List(context.TODO(), metav1.ListOptions{})
-		ctx.Ctx().Logf("%d %s remaining", len(res.Items), "subs")
-		return len(res.Items), err
-	})
-	Expect(err).NotTo(HaveOccurred())
+	Eventually(func() (remaining []v1alpha1.CatalogSource, err error) {
+		list := &v1alpha1.CatalogSourceList{}
+		err = client.List(clientCtx, list, inNamespace, ephemeralCatalogFieldSelector)
+		if list != nil {
+			remaining = list.Items
+		}
 
-	err = waitForEmptyList(func() (int, error) {
-		res, err := crc.OperatorsV1alpha1().CatalogSources(namespace).List(context.TODO(), metav1.ListOptions{FieldSelector: nonPersistentCatalogsFieldSelector})
-		ctx.Ctx().Logf("%d %s remaining", len(res.Items), "catalogs")
-		return len(res.Items), err
-	})
-	Expect(err).NotTo(HaveOccurred())
+		return
+	}).Should(BeEmpty(), "failed to await deletion of test catalogsources")
+
+	logf("deleting test crds...")
+	remainingCSVs := func() (csvs []v1alpha1.ClusterServiceVersion, err error) {
+		list := &v1alpha1.ClusterServiceVersionList{}
+		err = client.List(clientCtx, list, inNamespace, ephemeralCSVFieldSelector)
+		if list != nil {
+			csvs = list.Items
+		}
+
+		return
+	}
+
+	var crds []apiextensionsv1.CustomResourceDefinition
+	Eventually(func() error {
+		csvs, err := remainingCSVs()
+		if err != nil {
+			return err
+		}
+
+		for _, csv := range csvs {
+			for _, desc := range csv.Spec.CustomResourceDefinitions.Owned {
+				crd := &apiextensionsv1.CustomResourceDefinition{}
+				err := client.Get(clientCtx, types.NamespacedName{Name: desc.Name}, crd)
+				if apierrors.IsNotFound(err) {
+					continue
+				}
+				if err != nil {
+					return err
+				}
+				crds = append(crds, *crd)
+			}
+		}
+
+		return nil
+	}).Should(Succeed(), "failed to aggregate test crds for deletion")
+
+	Eventually(func() error {
+		for _, crd := range crds {
+			// Note: NotFound errors will be masked, so we can simply iterate until no other errors are returned.
+			// This is pretty inefficient, so if we're concerned about the number of API calls, we should replace this with something more sparing.
+			if err := client.Delete(clientCtx, &crd); MaskNotFound(err) != nil {
+				return err
+			}
+		}
+
+		return nil
+	}).Should(Succeed(), "failed to delete test crds")
+
+	logf("deleting test csvs...")
+	Eventually(func() error {
+		return client.DeleteAllOf(clientCtx, &v1alpha1.ClusterServiceVersion{}, inNamespace, ephemeralCSVFieldSelector)
+	}).Should(Succeed(), "failed to delete test csvs")
+
+	Eventually(remainingCSVs).Should(BeEmpty(), "failed to await deletion of test csvs")
+
+	logf("test resources deleted")
 }
 
 func buildCatalogSourceCleanupFunc(crc versioned.Interface, namespace string, catalogSource *v1alpha1.CatalogSource) cleanupFunc {
@@ -495,6 +555,43 @@ func createInternalCatalogSource(c operatorclient.ClientInterface, crc versioned
 	return catalogSource, cleanupInternalCatalogSource
 }
 
+func createInternalCatalogSourceWithPriority(c operatorclient.ClientInterface, crc versioned.Interface, name,
+	namespace string, manifests []registry.PackageManifest, crds []apiextensions.CustomResourceDefinition,
+	csvs []v1alpha1.ClusterServiceVersion, priority int) (*v1alpha1.CatalogSource, cleanupFunc) {
+
+	configMap, configMapCleanup := createConfigMapForCatalogData(c, name, namespace, manifests, crds, csvs)
+	// Create an internal CatalogSource custom resource pointing to the ConfigMap
+	catalogSource := &v1alpha1.CatalogSource{
+		TypeMeta: metav1.TypeMeta{
+			Kind:       v1alpha1.CatalogSourceKind,
+			APIVersion: v1alpha1.CatalogSourceCRDAPIVersion,
+		},
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      name,
+			Namespace: namespace,
+		},
+		Spec: v1alpha1.CatalogSourceSpec{
+			SourceType: "internal",
+			ConfigMap:  configMap.GetName(),
+			Priority:   priority,
+		},
+	}
+	catalogSource.SetNamespace(namespace)
+
+	ctx.Ctx().Logf("Creating catalog source %s in namespace %s...", name, namespace)
+	catalogSource, err := crc.OperatorsV1alpha1().CatalogSources(namespace).Create(context.TODO(), catalogSource, metav1.CreateOptions{})
+	if err != nil && !apierrors.IsAlreadyExists(err) {
+		Expect(err).ToNot(HaveOccurred())
+	}
+	ctx.Ctx().Logf("Catalog source %s created", name)
+
+	cleanupInternalCatalogSource := func() {
+		configMapCleanup()
+		buildCatalogSourceCleanupFunc(crc, namespace, catalogSource)()
+	}
+	return catalogSource, cleanupInternalCatalogSource
+}
+
 func createV1CRDInternalCatalogSource(t GinkgoTInterface, c operatorclient.ClientInterface, crc versioned.Interface, name, namespace string, manifests []registry.PackageManifest, crds []apiextensionsv1.CustomResourceDefinition, csvs []v1alpha1.ClusterServiceVersion) (*v1alpha1.CatalogSource, cleanupFunc) {
 	configMap, configMapCleanup := createV1CRDConfigMapForCatalogData(t, c, name, namespace, manifests, crds, csvs)
 
@@ -775,115 +872,6 @@ func deploymentReplicas(replicas int32) predicateFunc {
 	}), watch.Added, watch.Modified)
 }
 
-const (
-	cvoNamespace      = "openshift-cluster-version"
-	cvoDeploymentName = "cluster-version-operator"
-)
-
-func toggleCVO() {
-	c := ctx.Ctx().KubeClient().KubernetesInterface().AppsV1().Deployments(cvoNamespace)
-	clientCtx := context.Background()
-	scale, err := c.GetScale(clientCtx, cvoDeploymentName, metav1.GetOptions{})
-	if err != nil {
-		if apierrors.IsNotFound(err) {
-			// CVO is not enabled
-			return
-		}
-
-		Expect(err).ToNot(HaveOccurred())
-	}
-
-	if scale.Spec.Replicas > 0 {
-		scale.Spec.Replicas = 0
-	} else {
-		scale.Spec.Replicas = 1
-	}
-
-	Eventually(func() error {
-		_, err := c.UpdateScale(clientCtx, cvoDeploymentName, scale, metav1.UpdateOptions{})
-		return err
-	}).Should(Succeed())
-}
-
-// togglev2alpha1 toggles the v2alpha1 feature gate on or off.
-func togglev2alpha1() {
-	// Set the feature flag on OLM's deployment
-	c := ctx.Ctx().KubeClient()
-	deployment, err := getOperatorDeployment(c, operatorNamespace, labels.Set{"app": "olm-operator"})
-	Expect(err).ToNot(HaveOccurred())
-	toggleFeatureGates(deployment, feature.OperatorLifecycleManagerV2)
-}
-
-// toggleFeatureGates toggles the given feature gates on or off based on their current setting in the deployment.
-func toggleFeatureGates(deployment *appsv1.Deployment, toToggle ...featuregate.Feature) {
-	var (
-		c              = ctx.Ctx().KubeClient().KubernetesInterface().AppsV1().Deployments(deployment.GetNamespace())
-		containers     = deployment.Spec.Template.Spec.Containers
-		containerIndex = -1
-		argIndex       = -1
-		prefix         = "--feature-gates="
-		gateVals       string
-	)
-
-	// Find the container and argument indices for the feature gate option
-	for i, container := range containers {
-		if container.Name != "olm-operator" {
-			continue
-		}
-		containerIndex = i
-
-		for j, arg := range container.Args {
-			if gateVals = strings.TrimPrefix(arg, prefix); arg == gateVals {
-				continue
-			}
-			argIndex = j
-
-			break
-		}
-
-		break
-	}
-	// This should never happen since Deployments must have at least one container
-	Expect(containerIndex).ToNot(BeNumerically("<", 0), "deployment %s has no containers", deployment.GetName())
-
-	gate := feature.Gate.DeepCopy()
-	if argIndex >= 0 {
-		// Collect existing gate values
-		Expect(gate.Set(gateVals)).To(Succeed())
-	}
-
-	// Toggle gates
-	toggled := map[string]bool{}
-	for _, feature := range toToggle {
-		toggled[string(feature)] = !gate.Enabled(feature)
-	}
-	Expect(gate.SetFromMap(toggled)).To(Succeed())
-
-	gateArg := fmt.Sprintf("%s%s", prefix, gate)
-	if argIndex >= 0 {
-		// Overwrite existing gate options
-		containers[containerIndex].Args[argIndex] = gateArg
-	} else {
-		// No existing gate options, add one
-		containers[containerIndex].Args = append(containers[containerIndex].Args, gateArg)
-	}
-
-	clientCtx := context.Background()
-	w, err := c.Watch(clientCtx, metav1.ListOptions{})
-	Expect(err).ToNot(HaveOccurred())
-
-	timeout := 1 * time.Minute
-	Eventually(func() error {
-		_, err := c.Update(clientCtx, deployment, metav1.UpdateOptions{})
-		return err
-	}, timeout).Should(Succeed())
-
-	deadline, cancel := context.WithTimeout(context.Background(), timeout)
-	defer cancel()
-
-	awaitPredicates(deadline, w, deploymentReplicas(2), deploymentAvailable, deploymentReplicas(1))
-}
-
 func Apply(obj controllerclient.Object, changeFunc interface{}) func() error {
 	return ctx.Ctx().SSAClient().Apply(context.Background(), obj, changeFunc)
 }
